/*************************************************************************
* (C) 2013 AITNET ltd - Sofia/Bulgaria - <misho@aitnet.org>
*  by Michael Pounov <misho@elwix.org>
*
* $Author: misho $
* $Id: defs.h,v 1.4 2015/06/25 17:53:49 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2015
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#ifndef __DEFS_H
#define __DEFS_H


/* memory pool helper macros */

#define MEM_MAGIC_START	0xAAAAAAAA
#define MEM_MAGIC_STOP	0x55555555
#define MEM_ALLOC_MAX	0xFFFFFFF3

#define MEM_MIN_BUCKET	4
#define MEM_MAX_BUCKET	32

#define MEM_PTR2LINT(x)	((u_int) ((uintptr_t)(x)) & 0xFFFFFFFF)
#define MEM_PTR2HINT(x)	((u_int) (sizeof(uintptr_t) == sizeof(u_int)) ? 0 : \
				(((uintptr_t)(x)) >> 32) & 0xFFFFFFFF)
#define MEM_INT2PTR(x)	((void*) ((uintptr_t)(x)))

#define MEM_BADADDR(x)	(MEM_PTR2LINT(x) < 0x00001000U || MEM_PTR2LINT(x) > 0xFFFFF000U)
#define MEM_CORRUPT(x)	(!(x) || ((u_int*)(x))[-1] != MEM_MAGIC_START || \
				((u_int*)(x))[((u_int*)(x))[-2]] != MEM_MAGIC_STOP)

#define mem_data(x, _type)	(assert((x)), (_type) ((x)->alloc_mem ? ((x)->alloc_mem + 2) : NULL))
#define mem_len(x)		(assert((x)), (x)->alloc_mem ? *(x)->alloc_mem : 0)
#define mem_size(x)		(mem_len((x)) * sizeof(u_int))


/* CRC helpers */

#define MAX_FLETCHER16_DIGEST	21
#define MAX_FLETCHER_DIGEST	360
#define MAX_ADLER_DIGEST	5552

#define POLY_CRC32		0xC0A0A0D5
#define INIT_BERNSTEIN		0x1505
#define INIT_FNV1		0x811c9dc5
#define POLY_FNV1		0x1000193
#define INIT_JENKINS32		0xDEADBEEF

#define ROT32(x, k)		(((x) << (k)) | ((x) >> (32 - (k))))
#define MIX32(a, b, c) \
do { \
	  a -= c;  a ^= ROT32(c, 4);  c += b; \
	  b -= a;  b ^= ROT32(a, 6);  a += c; \
	  c -= b;  c ^= ROT32(b, 8);  b += a; \
	  a -= c;  a ^= ROT32(c, 16); c += b; \
	  b -= a;  b ^= ROT32(a, 19); a += c; \
	  c -= b;  c ^= ROT32(b, 4);  b += a; \
} while (0)
#define FINAL32(a, b, c) \
do { \
	  c ^= b; c -= ROT32(b, 14); \
	  a ^= c; a -= ROT32(c, 11); \
	  b ^= a; b -= ROT32(a, 25); \
	  c ^= b; c -= ROT32(b, 16); \
	  a ^= c; a -= ROT32(c, 4);  \
	  b ^= a; b -= ROT32(a, 14); \
	  c ^= b; c -= ROT32(b, 24); \
} while (0)


#define LOGERR	{ \
			elwix_Errno = errno; \
			strlcpy(elwix_Error, strerror(errno), sizeof elwix_Error); \
		}

extern mpool_t *elwix_mpool;
extern int use_mm;
extern const char elwix_Prog[STRSIZ];

extern int elwix_Errno;
extern char elwix_Error[STRSIZ];

void elwix_SetErr(int eno, char *estr, ...);


#endif
